# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='FailedLoginLog',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, auto_created=True, verbose_name='ID')),
                ('username', models.CharField(blank=True, null=True, max_length=255)),
                ('ip_address', models.CharField(blank=True, null=True, max_length=40, verbose_name='IP')),
                ('forwarded_by', models.CharField(blank=True, null=True, max_length=1000)),
                ('user_agent', models.CharField(blank=True, null=True, max_length=255)),
                ('timestamp', models.DateTimeField(auto_now_add=True)),
            ],
            options={
                'abstract': False,
                'ordering': ['-timestamp'],
            },
        ),
        migrations.CreateModel(
            name='LoginLog',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, auto_created=True, verbose_name='ID')),
                ('username', models.CharField(blank=True, null=True, max_length=255)),
                ('ip_address', models.CharField(blank=True, null=True, max_length=40, verbose_name='IP')),
                ('forwarded_by', models.CharField(blank=True, null=True, max_length=1000)),
                ('user_agent', models.CharField(blank=True, null=True, max_length=255)),
                ('timestamp', models.DateTimeField(auto_now_add=True)),
            ],
            options={
                'abstract': False,
                'ordering': ['-timestamp'],
            },
        ),
    ]



from django.contrib.auth import get_user_model
from django.contrib import admin
from django.urls import reverse
from django.utils.html import format_html

from useraudit import models as m


class LogAdmin(admin.ModelAdmin):
    model = m.Log

    search_fields = ['username']
    list_filter = ['timestamp']
    list_display = ('username', 'ip_address', 'forwarded_by', 'user_agent', 'timestamp')
    list_display_links = None


class LoginAttemptAdmin(admin.ModelAdmin):
    model = m.LoginAttempt

    list_display = ('username', 'count', 'timestamp', 'activate')
    list_display_links = None

    def activate(self, obj):
        UserModel = get_user_model()
        try:
            user = UserModel._default_manager.get_by_natural_key(obj.username)
            if user.is_active:
                return "Active"
            activation_url = reverse("useraudit:reactivate_user", args=[user.id])
            return format_html(u"<a href='%s'>Activate</a>" % activation_url)
        except UserModel.DoesNotExist:
            return "N/A"

    activate.short_description = "Status"
    activate.allow_tags = True


admin.site.register(m.LoginLog, LogAdmin)
admin.site.register(m.FailedLoginLog, LogAdmin)
admin.site.register(m.LoginAttempt, LoginAttemptAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('useraudit', '0003_auto_20160406_1434'),
    ]

    operations = [
        migrations.AlterField(
            model_name='failedloginlog',
            name='user_agent',
            field=models.CharField(max_length=1000, null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='loginlog',
            name='user_agent',
            field=models.CharField(max_length=1000, null=True, blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('useraudit', '0005_add_userdeactivation_table'),
    ]

    operations = [
        migrations.AlterField(
            model_name='userdeactivation',
            name='reason',
            field=models.CharField(null=True, max_length=2, blank=True, choices=[('AE', 'Account expired'), ('PE', 'Password expired'), ('FL', 'Too many failed login attemtps')]),
        ),
    ]



from django.dispatch import Signal

password_will_expire_warning = Signal(providing_args=["user", "days_left"])
password_has_expired = Signal(providing_args=["user"])
account_has_expired = Signal(providing_args=["user"])
login_failure_limit_reached = Signal(providing_args=["user"])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('useraudit', '0002_loginattempt'),
    ]

    operations = [
        migrations.AlterField(
            model_name='loginattempt',
            name='count',
            field=models.PositiveIntegerField(default=0, null=True, blank=True),
        ),
    ]



from __future__ import unicode_literals
import datetime
import logging
from django.db import models
from django.contrib.auth.signals import user_logged_in
from .signals import password_has_expired, account_has_expired, login_failure_limit_reached


logger = logging.getLogger('django.security')


class LoginAttempt(models.Model):
    username = models.CharField(max_length=255, null=True, blank=True)
    count = models.PositiveIntegerField(null=True, blank=True, default=0)
    timestamp = models.DateTimeField(auto_now_add=True)


class LoginAttemptLogger(object):

    def reset(self, username):
        defaults = {
            'count': 0,
            'timestamp': datetime.datetime.now()
        }
        LoginAttempt.objects.update_or_create(username=username, defaults=defaults)

    def increment(self, username):
        obj, created = LoginAttempt.objects.get_or_create(username=username)
        obj.count += 1
        obj.timestamp = datetime.datetime.now()
        obj.save()


class Log(models.Model):
    class Meta:
        abstract = True
        ordering = ['-timestamp']

    username = models.CharField(max_length=255, null=True, blank=True)
    ip_address = models.CharField(max_length=40, null=True, blank=True, verbose_name="IP")
    forwarded_by = models.CharField(max_length=1000, null=True, blank=True)
    user_agent = models.CharField(max_length=1000, null=True, blank=True)
    timestamp = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return '%s|%s|%s|%s|%s' % (self.username, self.ip_address, self.forwarded_by, self.user_agent, self.timestamp)


class UserDeactivation(models.Model):
    ACCOUNT_EXPIRED = 'AE'
    PASSWORD_EXPIRED = 'PE'
    TOO_MANY_FAILED_LOGINS = 'FL'

    DEACTIVATION_REASON_CHOICES = (
        (ACCOUNT_EXPIRED, 'Account expired'),
        (PASSWORD_EXPIRED, 'Password expired'),
        (TOO_MANY_FAILED_LOGINS, 'Too many failed login attempts'),
    )

    username = models.CharField(max_length=255)
    reason = models.CharField(max_length=2, blank=True, null=True, choices=DEACTIVATION_REASON_CHOICES)
    timestamp = models.DateTimeField(auto_now_add=True)


class FailedLoginLog(Log):
    pass


class LoginLog(Log):
    pass


class LoginLogger(object):

    def log_failed_login(self, username, request):
        fields = self.extract_log_info(username, request)
        FailedLoginLog.objects.create(**fields)

    def log_login(self, username, request):
        fields = self.extract_log_info(username, request)
        LoginLog.objects.create(**fields)

    def extract_log_info(self, username, request):
        USER_AGENT_MAX_LENGTH = Log._meta.get_field('user_agent').max_length
        if request:
            ip_address, proxies = self.extract_ip_address(request)
            user_agent = request.META.get('HTTP_USER_AGENT')
        else:
            ip_address = None
            proxies = None
            user_agent = None

        if user_agent is not None and len(user_agent) > USER_AGENT_MAX_LENGTH:
            logger.warning('Truncating User Agent to fit into %d. Original was: "%s"',
                           USER_AGENT_MAX_LENGTH, user_agent)
            user_agent = user_agent[:USER_AGENT_MAX_LENGTH]

        return {
            'username': username,
            'ip_address': ip_address,
            'user_agent': user_agent,
            'forwarded_by': ",".join(proxies or [])
        }

    def extract_ip_address(self, request):
        client_ip = request.META.get('REMOTE_ADDR')
        proxies = None
        forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
        if forwarded_for is not None:
            closest_proxy = client_ip
            forwarded_for_ips = [ip.strip() for ip in forwarded_for.split(',')]
            client_ip = forwarded_for_ips.pop(0)
            forwarded_for_ips.reverse()
            proxies = [closest_proxy] + forwarded_for_ips

        return (client_ip, proxies)


login_logger = LoginLogger()
login_attempt_logger = LoginAttemptLogger()


def login_callback(sender, user, request, **kwargs):
    username = user.get_username()
    login_logger.log_login(username, request)
    login_attempt_logger.reset(username)
    UserDeactivation.objects.filter(username=username).delete()


# User logged in Django signal
user_logged_in.connect(login_callback)


def save_login_deactivation(reason):
    def callback(sender, user, **kwargs):
        username = user.get_username()
        UserDeactivation.objects.filter(username=username).delete()
        UserDeactivation.objects.create(username=username, reason=reason)
    return callback


password_expired_callback = save_login_deactivation(UserDeactivation.PASSWORD_EXPIRED)
account_expired_callback = save_login_deactivation(UserDeactivation.ACCOUNT_EXPIRED)
login_failure_limit_reached_callback = save_login_deactivation(UserDeactivation.TOO_MANY_FAILED_LOGINS)


password_has_expired.connect(password_expired_callback)
account_has_expired.connect(account_expired_callback)
login_failure_limit_reached.connect(login_failure_limit_reached_callback)


# Import password expiry module so that the signal is registered.
# The password expiry feature won't be active unless the necessary
# settings are present.
from . import password_expiry  # noqa



import threading
try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    class MiddlewareMixin(object):
        def __init__(self, *args, **kwargs):
            pass

thread_data = threading.local()


def get_request():
    return getattr(thread_data, 'request', None)


class RequestToThreadLocalMiddleware(MiddlewareMixin):

    def process_request(self, request):
        thread_data.request = request



"""
Django password and account expiry.

This will prevent users from logging in unless they have changed their
password within a configurable password expiry period.

Expired users can reset their password using the normal registration
forms.

It will disable unused accounts. If users haven't logged in for a
certain time period, their account will be disabled next time a login
is attemped.

Requirement for account expiry: whichever user model is used should
implement AbstractBaseUser (standard Django user model does of
course).


How to use:

1. Add "useraudit" to the list of INSTALLED_APPS.

2. Put expiry backend *first* in the list of auth backends::

       AUTHENTICATION_BACKENDS = (
           'useraudit.password_expiry.AccountExpiryBackend',
           # ... the rest ...
       )

3. Use either Option A or Option B depending on your app.
    - A: Django custom auth model
    - B: "Profile" model related OneToOne with stock django User

   Option A: Use a django custom auth model for your users and add a field for
   password expiry::

       # settings.py
       AUTH_USER_MODEL = "myapp.MyUser"
       AUTH_USER_MODEL_PASSWORD_CHANGE_DATE_ATTR = "password_change_date"

       # models.py
       from django.contrib.auth.models import AbstractUser

       class MyUser(AbstractUser):
           password_change_date = models.DateTimeField(
               auto_now_add=True,
               null=True,
           )

   Option B: Use your existing profile model or create a new one, and
   add a field for password expiry::

       # settings.py
       AUTH_USER_MODEL_PASSWORD_CHANGE_DATE_ATTR = "myprofile.password_change_date"

       # models.py
       from django.db import models
       from django.contrib.auth.models import User

       class MyProfile(models.Model):
           user = models.OneToOneField(User, on_delete=models.CASCADE)
           password_change_date = models.DateTimeField(
               auto_now_add=True,
               null=True,
           )


4. Configure the settings relevant to password expiry::

       # How long a user's password is good for. None or 0 means no expiration.
       PASSWORD_EXPIRY_DAYS = 180
       # How long before expiry will the frontend start bothering the user
       PASSWORD_EXPIRY_WARNING_DAYS = 30
       # # Disable the user's account if they haven't logged in for this time
       # ACCOUNT_EXPIRY_DAYS = 100

5. Add log handlers for "django.security" if they aren't already there.

6. Inspect all non-standard login views and make sure they are
   checking for User.is_active.

7. Add code to your frontend to nag the user if their password is due
   to expire. Otherwise one day they will be unable to login and they
   won't know why.

   todo: add an automatic process for e-mailing users before password expiry

8. In your deployment scripts, include a daily cronjob to run the
   disable_inactive_users management command. This will let users know
   if their account has been disabled. It requires the Sites framework
   to be enabled, and for the user model to have an "email" attribute.
"""

from collections import namedtuple
from functools import reduce
from datetime import timedelta
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.db.models.signals import pre_save
from django.dispatch import receiver
from django.utils import timezone
import logging
from .backend import AuthFailedLoggerBackend
from .signals import password_has_expired, password_will_expire_warning, account_has_expired

logger = logging.getLogger("django.security")

__all__ = ["AccountExpiryBackend"]


@receiver(pre_save, sender=settings.AUTH_USER_MODEL)
def user_pre_save(sender, instance=None, raw=False, **kwargs):
    user = instance
    attrs = ExpirySettings.get()
    # We're saving the password change date only for existing users
    # Users just created should be taken care of by auto_now_add.
    # This way we can assume that a User profile object already exists
    # for the user. This is essential, because password change detection
    # can happen only in pre_save, in post_save it is too late.
    is_new_user = user.pk is None
    if is_new_user or raw:
        return
    if attrs.date_changed:
        update_date_changed(user, attrs.date_changed)

    # User has been re-activated. Ensure the last_login is set to None so
    # that the user isn't inactivated on next login by the AccountExpiryBackend
    current_user = sender.objects.get(pk=user.pk)
    if not current_user.is_active and user.is_active:
        user.last_login = None


def update_date_changed(user, date_changed_attr):
    def did_password_change(user):
        current_user = get_user_model().objects.get(pk=user.pk)
        return current_user.password != user.password

    def save_profile_password_change_date(user, date):
        parts = date_changed_attr.split('.')
        attr_name = parts[-1]
        profile = reduce(lambda obj, attr: getattr(obj, attr), parts[:-1], user)
        setattr(profile, attr_name, date)
        profile.save()

    def set_password_change_date(user, date):
        setattr(user, date_changed_attr, date)

    if did_password_change(user):
        now = timezone.now()
        if '.' in date_changed_attr:
            save_profile_password_change_date(user, now)
        else:
            set_password_change_date(user, now)


def should_warn_about_password_expiry(user):
    warn_in_days = ExpirySettings.get().num_warning_days or 0
    days_left = days_to_password_expiry(user) or 0
    return warn_in_days > 0 and days_left <= warn_in_days


def days_to_password_expiry(user):
    earliest = ExpirySettings.get().earliest_possible_password_change
    if earliest:
        change_date = get_password_change_date(user)
        if change_date:
            return (change_date - earliest).days


def is_password_expired(user):
    days = days_to_password_expiry(user)
    return days is not None and days < 0


def get_password_change_date(user):
    attr = ExpirySettings.get().date_changed
    if attr:
        val = user
        if isinstance(attr, str):
            for part in attr.split("."):
                if hasattr(val, part):
                    val = getattr(val, part)
                else:
                    logger.warning("User model does not have a %s attribute" % attr)
                    return None
            return val
        else:
            logger.warning("Password change attr in settings is not a string")


def get_user_last_login(user):
    if hasattr(user, "last_login"):
        return user.last_login
    else:
        logger.warning("User model doesn't have last_login field. ACCOUNT_EXPIRY_DAYS setting will have no effect.")
        return None


def is_account_expired(user):
    earliest = ExpirySettings.get().earliest_possible_login
    if earliest:
        last_login = get_user_last_login(user)
        return last_login and last_login < earliest
    return False


class ExpirySettings(namedtuple("ExpirySettings",
                                ["num_days", "num_warning_days", "date_changed", "password", "account_expiry"])):
    @classmethod
    def get(cls):
        expiry = getattr(settings, "PASSWORD_EXPIRY_DAYS", None) or 0
        warning = getattr(settings, "PASSWORD_EXPIRY_WARNING_DAYS", None) or 0
        date_changed = getattr(settings, "AUTH_USER_MODEL_PASSWORD_CHANGE_DATE_ATTR", None) or None
        password = getattr(settings, "AUTH_USER_MODEL_PASSWORD_ATTR", None) or "password"
        account_expiry = getattr(settings, "ACCOUNT_EXPIRY_DAYS", None) or 0
        return cls(expiry, warning, date_changed, password, account_expiry)

    @property
    def earliest_possible_login(self):
        if self.account_expiry > 0:
            return timezone.now() - timedelta(days=self.account_expiry)
        return None

    @property
    def earliest_possible_password_change(self):
        if self.num_days > 0:
            return timezone.now() - timedelta(days=self.num_days)
        return None


class AccountExpiryBackend(object):
    """
    This backend doesn't authenticate, it just prevents authentication
    of a user whose account password has expired.
    """
    def authenticate(self, request=None, username=None, password=None, **kwargs):
        user = self._lookup_user(username, password, **kwargs)

        if user:
            # Prevent authentication of inactive users (if the user
            # model supports it). Django only checks is_active at the
            # login view level.
            if hasattr(user, "is_active") and not user.is_active:
                self._prevent_login(username, "Account is not active")

            if is_password_expired(user):
                logger.info("Password expired! Disabling user account: %s" % user)
                user.is_active = False
                user.save()
                password_has_expired.send(sender=user.__class__, user=user)
                self._prevent_login(username, "Password has expired")

            if is_account_expired(user):
                logger.info("Disabling stale user account: %s" % user)
                user.is_active = False
                user.save()
                account_has_expired.send(sender=user.__class__, user=user)
                self._prevent_login(username, "Account has expired")

            if should_warn_about_password_expiry(user):
                days_left = days_to_password_expiry(user)
                logger.info("User's '%s' password will expire in %d days", user, days_left)
                password_will_expire_warning.send(sender=user.__class__, user=user, days_left=days_left)

        # pass on to next handler
        return None

    def _prevent_login(self, username, msg="User login prevented"):
        def is_failed_login_logger_configured():
            auth_backends = getattr(settings, 'AUTHENTICATION_BACKENDS', [])
            return 'useraudit.backend.AuthFailedLoggerBackend' in auth_backends

        logger.info("Login Prevented for user '%s'! %s", username, msg)
        if is_failed_login_logger_configured():
            AuthFailedLoggerBackend().authenticate(username=username)
        raise PermissionDenied(msg)

    def _lookup_user(self, username=None, password=None, **kwargs):
        # This is the same procedure as in
        # django.contrib.auth.backends.ModelBackend, except without
        # the timing attack mitigation, because it doesn't take long
        # to check for expiry.
        UserModel = get_user_model()
        if username is None:
            username = kwargs.get(UserModel.USERNAME_FIELD)
        try:
            return UserModel._default_manager.get_by_natural_key(username)
        except UserModel.DoesNotExist:
            return None



import logging
from django.contrib.auth import get_user_model
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.dispatch import receiver
from django.db.models.signals import pre_save
from django.views.decorators.debug import sensitive_variables

from .signals import login_failure_limit_reached
from .models import LoginLogger, LoginAttempt
from .models import LoginAttemptLogger
from .middleware import get_request


logger = logging.getLogger("django.security")


@receiver(pre_save, sender=settings.AUTH_USER_MODEL)
def user_pre_save(sender, instance=None, raw=False, **kwargs):
    user = instance
    is_new_user = user.pk is None
    if is_new_user or raw:
        return

    # User has been re-activated. Ensure the failed login counter is set to 0 so
    # that the user isn't inactivated on next login by the AuthFailedLoggerBackend
    current_user = sender.objects.get(pk=user.pk)
    if not current_user.is_active and user.is_active:
        LoginAttemptLogger().reset(user.username)


class AuthFailedLoggerBackend(object):

    def __init__(self):
        self.login_logger = LoginLogger()
        self.login_failure_limit = getattr(settings, 'LOGIN_FAILURE_LIMIT', None) or 0
        self.login_attempt_logger = LoginAttemptLogger()

    @sensitive_variables('credentials')
    def authenticate(self, request=None, **credentials):
        if request is None:
            # TODO remove this when we don't support Django 1.10 anymore
            # request is passed to authenticate starting with Django 1.11
            request = get_request()
        UserModel = get_user_model()
        self.username = credentials.get(UserModel.USERNAME_FIELD)
        self.login_logger.log_failed_login(self.username, request)
        if self._get_user() is not None:
            self.login_attempt_logger.increment(self.username)
            self.block_user_if_needed()

        return None

    def block_user_if_needed(self):
        if not self.is_login_failure_limit_enabled():
            return
        if self.is_attempts_exceeded():
            self._deactivate_user()
            user = self._get_user()
            login_failure_limit_reached.send(sender=user.__class__, user=user)
            logger.info("Login Prevented for user '%s'! Maximum failed logins %d reached!",
                        self.username, self.login_failure_limit)
            raise PermissionDenied("Username '%s' has been blocked" % self.username)

    def is_login_failure_limit_enabled(self):
        return self.login_failure_limit > 0

    def is_attempts_exceeded(self,):
        count = self._get_count()
        if count and count >= self.login_failure_limit:
            return True
        return False

    def _get_count(self):
        try:
            obj = LoginAttempt.objects.get(username=self.username)
            return obj.count
        except LoginAttempt.DoesNotExist:
            return None

    def _get_user(self):
        UserModel = get_user_model()
        try:
            return UserModel._default_manager.get_by_natural_key(self.username)
        except UserModel.DoesNotExist:
            logger.warning("User model for username %s not found" % self.username)
            return None

    def _deactivate_user(self):
        user = self._get_user()
        if user:
            user.is_active = False
            user.save(update_fields=["is_active"])
            logger.warning("Username '%s' has been blocked" % self.username)
            return True
        return False



from django.contrib import admin
from .views import test_request_available

try:
    from django.urls import include, path
    PRE_DJANGO_2 = False
except ImportError:
    from django.conf.urls import include, url
    PRE_DJANGO_2 = True


admin.autodiscover()

if PRE_DJANGO_2:
    urlpatterns = [
        url(r'^admin/', include(admin.site.urls)),
        url(r'test_request_available[/]?$', test_request_available),
    ]
else:
    urlpatterns = [
        path('admin/', admin.site.urls),
        path('test_request_available/', test_request_available),
    ]



#!/usr/bin/env python

try:
    from setuptools.core import setup
except ImportError:
    from distutils.core import setup

setup(
    name='django-useraudit',
    version='1.7.1',
    description='Django user audit utilities like logging user log in, disabling access when password expires or user is inactive',
    long_description='Django user audit utilities like logging user log in, disabling access when password expires or user is inactive',
    author='CCG, Murdoch University',
    author_email='devops@ccg.murdoch.edu.au',
    url='https://github.com/muccg/django-useraudit',
    download_url='https://github.com/muccg/django-useraudit/releases',
    classifiers=[
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development"
    ],
    zip_safe=True,
    packages=[
        'useraudit',
        'useraudit.migrations',
    ],
    include_package_date=True,
)



import logging

from django.http import HttpResponse, HttpResponseNotFound, HttpResponseRedirect
from django.contrib.auth import get_user_model
from django.urls import reverse

from .models import LoginAttemptLogger
from . import middleware


logger = logging.getLogger("django.security")

login_attempt_logger = LoginAttemptLogger()


def test_request_available(request):
    thread_request = middleware.get_request()
    if thread_request == request:
        return HttpResponse('OK')
    return HttpResponseNotFound()


def reactivate_user(request, user_id):
    user = _get_user(user_id)
    user.is_active = True
    user.save()
    login_attempt_logger.reset(user.username)
    return HttpResponseRedirect(reverse("admin:useraudit_loginattempt_changelist"))


def _get_user(user_id):
    UserModel = get_user_model()
    try:
        return UserModel.objects.get(id=user_id)
    except UserModel.DoesNotExist:
        logger.warning("User model for user_id %d not found" % user_id)
        return None



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('useraudit', '0004_enlarge_user_agent_field'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserDeactivation',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('username', models.CharField(max_length=255)),
                ('reason', models.CharField(blank=True, max_length=2, null=True, choices=[(b'AE', b'Account expired'), (b'PE', b'Password expired'), (b'FL', b'Too many failed login attemtps')])),
                ('timestamp', models.DateTimeField(auto_now_add=True)),
            ],
        ),
    ]
